# Astro SDK Extras project
# (c) kol, 2023-2024

""" Custom PostgreSQL table merge operator. Based on Astro-SDK PostgresDatabase.merge_table() """

import logging
from psycopg2 import sql as postgres_sql
from sqlalchemy import text
from sqlalchemy.engine.base import Connection as SqlaConnection
from astro.constants import MergeConflictStrategy
from astro.table import BaseTable

_logger = logging.getLogger('airflow.task')

def postgres_merge_tables(
        conn: SqlaConnection,
        source_table: BaseTable | None,
        target_table: BaseTable,
        source_to_target_columns_map: dict[str, str],
        target_conflict_columns: list[str],
        source_sql: str | None = None,
        if_conflicts: MergeConflictStrategy = "update",
) -> int:
    """
    Merge the source table rows into a destination table.
    The argument `if_conflicts` allows the user to define how to handle conflicts.

    Differs from Astro-SDK `PostgresDatabase.merge_table()` as it supports pre-established
    connections (probably within transaction), custom `select` as data source
    and in that it will return number of rows affected.

    :param conn: Connection to execute merge on.
    :param source_table: Contains the rows to be merged to the target_table. Can be `None`.
    :param target_table: Contains the destination table in which the rows will be merged
    :param source_to_target_columns_map: Dict of target_table columns names to source_table columns names
    :param target_conflict_columns: List of cols where we expect to have a conflict while combining
    :param source_sql:   Query to get rows to be merged. Either this or `source_table` has to be provided.
    :param if_conflicts: The strategy to be applied if there are conflicts.
    """

    def identifier_args(table: BaseTable):
        schema = table.metadata.schema
        return (schema, table.name) if schema else (table.name,)

    if source_table is None and not source_sql:
        raise ValueError('Either `source_table` or `source_sql` must be provided')

    statement = (
        "INSERT INTO {target_table} ({target_columns}) SELECT {source_columns} FROM {source_table}"
    )

    source_columns = list(source_to_target_columns_map.keys())
    target_columns = list(source_to_target_columns_map.values())

    if if_conflicts == "ignore":
        statement += " ON CONFLICT ({target_conflict_columns}) DO NOTHING"
    elif if_conflicts == "update":
        statement += " ON CONFLICT ({target_conflict_columns}) DO UPDATE SET {update_statements}"

    source_column_names = [postgres_sql.Identifier(col) for col in source_columns]
    target_column_names = [postgres_sql.Identifier(col) for col in target_columns]
    update_statements = [
        postgres_sql.SQL("{col_name}=EXCLUDED.{col_name}").format(col_name=col_name) \
            for col_name in target_column_names
    ]

    query = postgres_sql.SQL(statement).format(
        target_columns=postgres_sql.SQL(",").join(target_column_names),
        target_table=postgres_sql.Identifier(*identifier_args(target_table)),
        source_columns=postgres_sql.SQL(",").join(source_column_names),
        source_table=postgres_sql.Identifier(*identifier_args(source_table)) if source_table \
                    else postgres_sql.SQL("(" + source_sql + ") q"),
        update_statements=postgres_sql.SQL(",").join(update_statements),
        target_conflict_columns=postgres_sql.SQL(",").join(
            [postgres_sql.Identifier(x) for x in target_conflict_columns]
        ),
    )

    sql = query.as_string(conn.connection.dbapi_connection)
    _logger.info(f'Executing {sql}')
    return conn.execute(text(sql).execution_options(autocommit=True)).rowcount


# Astro SDK Extras project
# (c) kol, 2023

""" Template utility functions """

import os
import jinja2
import importlib.resources
from airflow.models import DAG
from airflow.models.dag import DagContext
from airflow.configuration import conf
from airflow.exceptions import AirflowFailException

def get_template_file(
        template: str, 
        ext: str = '.sql', 
        dag: DAG | None = None,
        fail_if_not_found: bool = False) -> str | None:
    
    """ Returns reference to DAG-specific template file, if one exists.
    
    This function looks up for a template file in `./templates/<dag_id>` folder
    under directory set as DAGS_HOME in Airflow config (usually `./dags`).

    Templates are used for SQL operations, html reporting and so on.
    They consists of some plain-text content mixed with Jinja macros inside `{{...}}` brackets,
    which will be substituted ('rendered`) with some variables either by Airflow runtime or manually.

    Args:
        template:   Template name (e.g. table name or reporting object) without any extension
        ext:        Template file extension, `.sql` by default
        dag:        Optional DAG context. If not provided, current DAG context will be used
        fail_if_not_found: If `True`, `AirflowFailException` will be raised if template does not exist
            (default is `False`)

    Returns:
        Name of template file relative to Airflow's `DAG_HOME` folder or `None` if no file was found
        and `fail_if_not_found = False`

    Examples:
        If a `test_table` is to be processed in DAG with `test-transfer` ID, 
        this SQL template could be used to add required `session_id` field 
        at the beginning of a target table:

            select {{ti.xcom_pull(key="session").session_id}} as session_id, * 
            from test_table

        The template must be located in `./dags/templates/test-transfer/test_table.sql` file.

        Then, during this DAG execution, the template will automatically be picked up and executed,
        substituting macros with actual `session_id` value:

        >>> with DAG(dag_id='test-transfer', ...) as dag:
        >>>     session = open_session('source', 'target')
        >>>     transfer_table('test_table', session=session)
        >>>     close_session(session)
    """

    dag = dag or DagContext.get_current_dag()
    ext = '.' + ext if not ext.startswith('.') else ext
    rel_file_name = os.path.join('templates', dag.dag_id, template + ext)
    full_file_name = os.path.join(conf.get('core','dags_folder'), rel_file_name)
    if os.path.exists(full_file_name):
        return rel_file_name
    if fail_if_not_found:
        raise AirflowFailException(f'Template file {full_file_name} was not found')
    return None

def get_template(
        template: str, 
        ext: str = '.sql', 
        dag: DAG | None = None,
        fail_if_not_found: bool = False,
        read_mode: str = 'rt') -> str | bytes | None:
    
    """ Returns template content. See `get_template_file` for details on templates.

    Args:
        template:   Template name (e.g. table name or reporting object) without any extension
        ext:        Template file extension, `.sql` by default
        dag:        Optional DAG context. If not provided, current DAG context will be used
        fail_if_not_found: If `True`, `AirflowFailException` will be raised if template does not exist
        read_mode:  Read mode (either `rt` for text or `rb` for binary)

    Returns:
        Template file content or `None` if no file was found and `fail_if_not_found = False`.

    Examples:
        Load template from `./dags/templates/test_dag/mail.html`, manually render it
        with Airflow DAG context and send it as a mail:

        >>> from jinja2 import Template
        >>> from airflow.operators.email import EmailOperator
        >>> from airflow.operators.python import get_current_context
        >>>
        >>> with DAG(dag_id='test_dag', ...) as dag:
        >>>     @dag.task
        >>>     def get_content():
        >>>         templ = get_template('mail', '.html', dag=dag)
        >>>         return Template(templ).render(get_current_context())
        >>>
        >>>     EmailOperator(task_id='send_email', to='somebody@somewhere.com', 
        >>>         subject='Important!', html_content=get_content())

        Note that because `EmailOperator.html_content` is a templated field,
        Airflow will automatically render it, so this example might even more be shortened:

        >>> from airflow.operators.email import EmailOperator
        >>> from airflow.operators.python import get_current_context
        >>>
        >>> with DAG(dag_id='test_dag', ...) as dag:
        >>>     EmailOperator(task_id='send_email', to='somebody@somewhere.com', 
        >>>         subject='Important!', 
        >>>         html_content=get_template_file('mail', '.html'))
        
    """
    assert read_mode in ('rt', 'rb'), f'Read_mode must be either `rt` or `rb`'
    template_file = get_template_file(template, ext, dag, fail_if_not_found)
    if not template_file:
        return None
    full_file_name = os.path.join(conf.get('core','dags_folder'), template_file)
    with open(full_file_name, read_mode) as fp:
        return fp.read()

def get_predefined_template(name: str) -> jinja2.Template:
    """ Retrieves a template from package resources. Mostly for internal use. """
    template_file = f'templates/{name}'
    template_res = importlib.resources.files('astro_extras').joinpath(template_file)
    with importlib.resources.as_file(template_res) as template_file:
        return jinja2.Template(template_file.read_text())



# Astro SDK Extras project
# (c) kol, 2023

""" ETL session unit tests """

from dateutil.parser import isoparse
from confsupport import run_dag, logger
from sqlalchemy import text

def get_last_session(target_db):
    sql = 'select * from sessions where session_id = (select max(session_id) from sessions)'
    with target_db.begin() as conn:
        cursor = conn.execute(text(sql))
        return cursor.mappings().one_or_none()

def get_last_session_id(target_db):
    last_session = get_last_session(target_db)
    return last_session['session_id'] if last_session else 0

def compare_period(session_period, expected_period):
    p = tuple([t.replace(tzinfo=None) for t in session_period])
    x = tuple([isoparse(t) for t in expected_period])
    return p == x

def test_session_simple(docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing session opening')
    prev_session_id = get_last_session_id(target_db)
    result = run_dag('test-session', docker_ip, docker_services, airflow_credentials)
    assert result == 'success'
    new_session = get_last_session(target_db)
    assert new_session
    logger.info(f'Latest session: {new_session}')
    assert new_session['session_id'] > prev_session_id
    assert new_session['status'] == 'success'

def test_session_ctx(docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing session opening using context manager semantics')
    prev_session_id = get_last_session_id(target_db)
    result = run_dag('test-session-ctx', docker_ip, docker_services, airflow_credentials)
    assert result == 'success'
    new_session = get_last_session(target_db)
    assert new_session
    logger.info(f'Latest session: {new_session}')
    assert new_session['session_id'] > prev_session_id
    assert new_session['status'] == 'success'

def test_session_period(airflow, docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing manual period assignment')
    result = run_dag('test-session', docker_ip, docker_services, airflow_credentials,
                     params='{"period": "[2023-05-01, 2023-05-31]"}')
    assert result == 'success'
    new_session = get_last_session(target_db)
    assert new_session
    logger.info(f'Latest session: {new_session}')
    assert new_session['status'] == 'success'
    assert compare_period(new_session['period'], ['2023-05-01T00:00:00', '2023-06-01T00:00:00'])

def test_session_one_day_period(airflow, docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing manual one-day period assignment')
    result = run_dag('test-session', docker_ip, docker_services, airflow_credentials,
                     params='{"period": "[2023-05-01, 2023-05-01]"}')
    assert result == 'success'
    new_session = get_last_session(target_db)
    assert new_session
    logger.info(f'Latest session: {new_session}')
    assert new_session['status'] == 'success'
    assert compare_period(new_session['period'] ,['2023-05-01T00:00:00', '2023-05-02T00:00:00'])

def test_session_timestamp_period(airflow, docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing manual timestamp period assignment')
    result = run_dag('test-session', docker_ip, docker_services, airflow_credentials,
                     params='{"period": "[2023-05-01T12:00:00, 2023-05-01T14:00:00]"}')
    assert result == 'success'
    new_session = get_last_session(target_db)
    assert new_session
    logger.info(f'Latest session: {new_session}')
    assert new_session['status'] == 'success'
    assert compare_period(new_session['period'], ['2023-05-01T12:00:00', '2023-05-01T14:00:00'])

def test_session_wrong_period(airflow, docker_ip, docker_services, airflow_credentials, target_db):
    logger.info(f'Testing invalid periods assignment')
    for p in '[2023-05-02, 2023-05-01]', '[2023-05-02]', 'aaa':
        result = run_dag('test-session', docker_ip, docker_services, airflow_credentials,
                        params='{"period": "' + p + '"}')
        assert result == 'failed'



# Astro SDK Extras project
# (c) kol, 2023

""" Dataframe comparsion utility function """

import logging
import pandas as pd
from typing import Union, Iterable, Tuple

from airflow.settings import TIMEZONE
from airflow.exceptions import AirflowFailException

def _split_row(vals, id_cols, shared_cols):
    """ Internal - split row by subsets """
    split_vals = {'x': {}, 'y': {}, '': {}}

    for id_key in id_cols:
        if id_key in vals:
            split_vals['x'][id_key] = vals[id_key]
            split_vals['y'][id_key] = vals[id_key]

    for shared_key in shared_cols:
        if shared_key in vals:
            split_vals[''][shared_key] = vals[shared_key]
        for suffix in ('x', 'y'):
            key = f'{shared_key}_{suffix}'
            if key in vals:
                split_vals[suffix][shared_key] = vals[key]
    return split_vals['x'], split_vals['y']

def _adjust_timestamps(df: pd.DataFrame, exclude_cols: Iterable):
    """ Internal: adjust timestamps to common TZ """

    logger = logging.getLogger('airflow.task')
    tz_cols = [col for col in df.columns if df[col].dtype == 'datetime64[ns]' and
                col not in exclude_cols]
    if tz_cols:
        logger.debug(f'Converting TZ-unaware timestamp columns {tz_cols} to timezone {TIMEZONE.name}')
        for col in tz_cols:
            df[col] = df[col].map(lambda t: t.tz_localize(TIMEZONE.name))

    return df

def _check_timestamp_types(df_a: pd.DataFrame, df_b: pd.DataFrame, exclude_cols: Iterable):
    """ Internal: check timestamp compatibility """

    tz_a = [col for col in df_a.columns if 'datetime64[ns' in str(df_a.dtypes[col]) and
                col not in exclude_cols]
    tz_b = [col for col in df_b.columns if 'datetime64[ns' in str(df_b.dtypes[col]) and
                col not in exclude_cols]
    for col in set(tz_a).intersection(tz_b):
        if df_a.dtypes[col] == 'datetime64[ns]' and df_b.dtypes[col] != 'datetime64[ns]':
            raise AirflowFailException(f'Incompatible timestamp column {col} type: ' \
                                   f'{df_a.dtypes[col]} on source, {df_b.dtypes[col]} on target')

def compare_datasets(
    df_src: pd.DataFrame, 
    df_trg: pd.DataFrame, 
    id_cols: Iterable | None = None, 
    exclude_cols: Iterable | None = None, 
    stop_on_first_diff: bool = True,
    logger: logging.Logger | None = None) -> Union[bool, Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:

    """ Compare two dataframes by columns, their types and values.

    Args:
        df_src: Source dataframe

        df_trg: Target dataframe

        id_cols: List of columns treated as IDs.
            If empty, 1st column is considered to be ID

        exclude_cols: List of columns to be excluded from comparsion

        stop_on_first_diff: if `True`, any difference would stop the comparsion,
            otherwise it will continue and produce difference dataframes

        logger: logger object (if `None`, `airflow.task` logger will be used)

    Returns:
        If `stop_on_first_diff == True`, would return boolean flag whether
        the dataframes are the same (`True`) or not (`False`).

        Otherwise, will return 3 dataframes for new, changed and 
        deleted records. Any of them could be `None` meaning
        no corresponding difference detected.
    """

    logger = logger or logging.getLogger('airflow.task')

    orig_src_cols = list(df_src.columns)
    src_cols = df_src.columns.str.lower()
    trg_cols = df_trg.columns.str.lower()
    df_src.columns = src_cols
    df_trg.columns = trg_cols

    logger.debug(f'Source cols: {src_cols}')
    logger.debug(f'Target cols: {trg_cols}')

    if not exclude_cols:
        exclude_cols = []
    else:
        exclude_cols = [c.lower() for c in exclude_cols]
    exclude_cols.extend(['session_id', '_created', '_modified', '_deleted'])
    logger.debug(f'Excluded columns: {exclude_cols}')

    if not id_cols:
        id_cols = [c for c in src_cols if c not in exclude_cols][:1]
    else:
        id_cols = [c.lower() for c in id_cols]
        diff_cols = set(id_cols).difference(src_cols)
        if diff_cols:
            raise AirflowFailException(f'ID columns missing from source dataset: {diff_cols}')
        diff_cols = set(id_cols).difference(trg_cols)
        if diff_cols:
            raise AirflowFailException(f'ID columns missing from target dataset: {diff_cols}')
    logger.debug(f'ID columns: {id_cols}')

    diff_cols = set(src_cols).symmetric_difference(trg_cols).difference(exclude_cols)
    if diff_cols:
        logger.warning(f'Diff in columns detected: {diff_cols}.\n' + 
                       'Operation will continue, but may result in incomplete data transfer or errors.\n' +
                       'Review the datasets and correct structure to avoid that.')

    if df_src.shape[0] != df_trg.shape[0] and stop_on_first_diff:
        logger.info('Number of records in source and target datasets is different')
        return False

    _check_timestamp_types(df_src, df_trg, exclude_cols)

    shared_cols = set(src_cols).intersection(trg_cols).difference(exclude_cols)
    logger.debug(f'Shared cols: {shared_cols}')

    df_src.index.name = '__src_idx__'
    df_trg.index.name = '__trg_idx__'
    df_deleted = None
    df_merged = pd.merge(df_src.reset_index(), df_trg.reset_index(), how='outer', on=id_cols,
                            copy=False, indicator=True, sort=False)
    if df_merged['_merge'].eq('right_only').any():
        logger.info('Target dataset contains records not present on the source')
        if stop_on_first_diff:
            return False

        idx_list = df_merged[df_merged['_merge'] == 'right_only']['__trg_idx__']
        df_deleted = df_trg.iloc[idx_list]
        df_deleted.index.name = None

    def _iter_tuples():
        for row in df_merged.itertuples(index=False, name=None):
            d = {c: v for c, v in zip(df_merged.columns, row)}
            yield d.pop('__src_idx__'), d.pop('__trg_idx__'), d

    new_idx, upd_idx = [], []
    for src_idx, trg_idx, vals in _iter_tuples():
        src_vals, trg_vals = _split_row(vals, id_cols, shared_cols)
        key_val = [src_vals[c] for c in id_cols]

        logger.debug(f'Src {src_idx}: {src_vals}')
        logger.debug(f'Trg {trg_idx}: {trg_vals}')

        if not pd.isnull(vals.get('_deleted')) or vals.get('_merge') == 'right_only':
            logger.debug('Deleted record, skipping')
            continue
        if pd.isnull(trg_idx):
            if stop_on_first_diff:
                logger.info(f'New record at {key_val}')
                return False

            logger.debug(f'New record at {key_val}')
            new_idx.append(src_idx)
        else:
            diff_values = {k: {'source': src_vals[k], 'target': trg_vals[k]} for k in shared_cols \
                if not (pd.isnull(src_vals[k]) and pd.isnull(trg_vals[k])) \
                    and src_vals[k] != trg_vals[k]}
            if diff_values:
                key_val = [src_vals[c] for c in id_cols]
                if stop_on_first_diff:
                    logger.info(f'Diff at {id_cols}={key_val}: {diff_values}')
                    return False
                
                logger.debug(f'Diff at {id_cols}={key_val}: {diff_values}')
                upd_idx.append(src_idx)

    if stop_on_first_diff:
        logger.info(f'No differences detected')
        return True

    if not new_idx and not upd_idx and (df_deleted is None or df_deleted.empty):
        logger.info(f'No differences detected')
        return None, None, None

    if not new_idx:
        df_new = None
        logger.info(f'No new records on source')
    else:
        df_new = df_src.iloc[new_idx]
        df_new.index.name = None
        df_new.columns = orig_src_cols
        logger.info(f'{df_new.shape[0]} new records on source')

    if not upd_idx:
        df_upd = None
        logger.info(f'No modified records on source')
    else:
        df_upd = df_src.iloc[upd_idx]
        df_upd.index.name = None
        df_upd.columns = orig_src_cols
        logger.info(f'{df_upd.shape[0]} modified records on source')

    logger.info(f'{df_deleted.shape[0] if df_deleted is not None else 0} records deleted on target')

    return df_new, df_upd, df_deleted


def compare_timed_dict(df_src: pd.DataFrame, df_trg: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ Compares two datasets as if they are timed dictionary data. Returns dataframes with records to open and to close """

    # Check target table is really a timed dictionary
    if not all([c in df_trg.columns for c in ['src_id', 'effective_from', 'effective_to']]):
        raise AirflowFailException(f'Destination table must contain `src_id`, `effective_from` and `effective_to` columns')

    # 1st column after `session_id` in a source table is considered to be ID
    src_id_col = [c for c in df_src.columns if c != 'session_id'][0]
    df_src.rename(columns={src_id_col: 'src_id'}, inplace=True)

    # Get the deltas
    df_new, df_upd, df_deleted = \
        compare_datasets(df_src, df_trg,
                            id_cols=['src_id'],
                            exclude_cols=['effective_from', 'effective_to'],
                            stop_on_first_diff=False)

    # New records are only for opening new record
    # Modified records are closed with opening new record
    # Deleted records are closed without opening new record
    effective_from = pd.Timestamp.now().floor('S')
    effective_to = effective_from - pd.Timedelta(seconds=1)

    df_closing = None
    if df_upd is not None and df_deleted is not None:
        df_closing = pd.concat([df_upd, df_deleted], ignore_index=True)
    elif df_upd is None and df_deleted is not None:
        df_closing = df_deleted
    elif df_upd is not None and df_deleted is None:
        df_closing = df_upd.copy()

    if df_closing is not None:
        df_closing['effective_to'] = effective_to

    df_opening = None
    if df_new is not None and df_upd is not None:
        df_opening = pd.concat([df_new, df_upd], ignore_index=True)
    elif df_new is None and df_upd is not None:
        df_opening = df_upd.copy()
    elif df_new is not None and df_upd is None:
        df_opening = df_new

    if df_opening is not None:
        df_opening['effective_from'] = effective_from
        df_opening['effective_to'] = None

    return df_opening, df_closing



# Astro SDK Extras project
# (c) kol, 2023

""" ETL session management routines """

import re
from datetime import datetime, timedelta
from dateutil.parser import isoparse
from sqlalchemy import text
from airflow.models import DAG, BaseOperator
from airflow.models.xcom_arg import XComArg
from airflow.models.dag import DagContext
from airflow.operators.python import get_current_context
from airflow.utils.context import Context
from airflow.utils.state import TaskInstanceState
from airflow.utils.trigger_rule import TriggerRule
from airflow.providers.common.sql.hooks.sql import DbApiHook
from airflow.exceptions import AirflowException, AirflowFailException
from airflow.settings import TIMEZONE
from attr import define, field
from astro import sql as aql
from typing import Any, Tuple, cast

from ..utils.datetime_local import datetime_to_tz
from ..utils.template import get_predefined_template

@define(slots=False)
class ETLSession:
    """ Session data object. Holds all ETL session attributes, can be pushed to XCom.
    Implements context manager protocol (see examples). 
    
    See `open_session` function for more information.

    Args:
        source_conn_id:   Source connection ID
        destination_conn_id:   Destination connection ID
        session_conn_id:   ID of connection, where `public.sessions` table is located.
            If not set, then `destination_conn_id` is used
        session_id:   Actual session ID (automatically generated, do not set it manually)
        period_start: Date and time of session period start as ISO-format string.
            See `open_session` for details.
        period_end: Date and time of session period end as ISO-format string.
            See `open_session` for details.
        dag: DAG, where the session was created. Used only to pass DAG's reference throught.

    Examples:
        Using `ETLSession` as context manager:
        >>> with DAG(...) as dag, ETLSession('source', 'target') as session:
        >>>     transfer_table('test_table', session=session)

        Explicit call to `open_session`:        
        >>> with DAG(...) as dag:
        >>>     session = open_session('source', 'target')
        >>>     transfer_table('test_table', session=session)
        >>>     close_session(session)
    """
    source_conn_id: str = field(default=None)
    destination_conn_id: str = field(default=None)
    session_conn_id: str = field(default=None)
    session_id: int = field(default=0)
    period_start: str = field(default=None)
    period_end: str = field(default=None)
    dag: DAG = field(default=None)

    def __attrs_post_init__(self) -> None:
        if not self.session_conn_id:
            self.session_conn_id = self.destination_conn_id

    def __getstate__(self):
        return self.__dict__

    def serialize(self):
        return {
            'source_conn_id': self.source_conn_id,
            'destination_conn_id': self.destination_conn_id,
            'session_conn_id': self.session_conn_id,
            'session_id': self.session_id,
            'period_start': self.period_start,
            'period_end': self.period_end,
        }

    @staticmethod
    def deserialize(data, version: int):
        return ETLSession(
            source_conn_id=data['source_conn_id'],
            destination_conn_id=data['destination_conn_id'],
            session_conn_id=data['session_conn_id'],
            session_id=data['session_id'],
            period_start=data['period_start'],
            period_end=data['period_end'],
        )

    def __enter__(self):
        self._actual_sesssion = open_session(
            source_conn_id=self.source_conn_id,
            destination_conn_id=self.destination_conn_id,
            session_conn_id=self.session_conn_id,
            dag=self.dag)
        return self._actual_sesssion

    def __exit__(self, type, value, traceback):
        dag = cast(DAG, DagContext.get_current_dag())
        if len(dag.tasks) > 1:
            t1, t2 = dag.tasks[0], dag.tasks[1]
            if t1.task_id == 'open-session' and 'open-session' not in t2.upstream_task_ids:
                if t2.task_group is None or t2.task_group.is_root:
                    t2.set_upstream(t1)
                else:
                    t2.task_group.set_upstream(t1)

        close_session(self._actual_sesssion).set_downstream(aql.cleanup())

class OpenSessionOperator(BaseOperator):
    """ Session opening operator. Normally is used within `open_session` function """
    ui_color = "#82eef0"

    def __init__(self,
                 *,
                 source_conn_id: str = 'default',
                 destination_conn_id: str = 'default',
                 session_conn_id: str = None,
                 **kwargs):

        task_id = kwargs.pop('task_id', 'open-session')
        super().__init__(task_id=task_id, **kwargs)

        self.source_conn_id = source_conn_id
        self.destination_conn_id = destination_conn_id
        self.session_conn_id = session_conn_id if session_conn_id else destination_conn_id
        self.session: ETLSession = None

    def _new_session(self, period_start: str, period_end: str, context: Context) -> int:
        template = get_predefined_template('session_open.sql')
        sql = template.render(
            source_conn_id=self.source_conn_id,
            destination_conn_id=self.destination_conn_id,
            period_str=f'{{ "{period_start}", "{period_end}" }}',
            started=datetime.now().isoformat(),
            run_id=context['run_id'],
        )
        hook: DbApiHook = DbApiHook.get_hook(self.session_conn_id)
        result = hook.get_first(sql)
        return result[0]

    def execute(self, context: Context):
        period_start, period_end = get_session_period(context)
        session_id = self._new_session(period_start, period_end, context)
        self.log.info(f'New session {session_id} for period [{period_start},{period_end}] started')

        session = ETLSession(
            source_conn_id=self.source_conn_id,
            destination_conn_id=self.destination_conn_id, 
            session_id=session_id, 
            session_conn_id=self.session_conn_id,
            period_start=period_start,
            period_end=period_end)
        
        context['ti'].xcom_push(key='session', value=session)
        return session

class CloseSessionOperator(BaseOperator):
    """ Session closing operator. Normally is used within `close_session` function """
    ui_color = "#78f0f0"

    def __init__(self,
                 *,
                 session: ETLSession | XComArg | None = None,
                 **kwargs):

        task_id = kwargs.pop('task_id', 'close-session')
        trigger_rule = kwargs.pop('trigger_rule', TriggerRule.ALL_DONE)
        super().__init__(task_id=task_id, trigger_rule=trigger_rule, **kwargs)
        self.session = session

    def execute(self, context: Context):
        self.session = ensure_session(self.session)

        dag_run = context['dag_run']
        failed_tasks = [ti for ti in dag_run.get_task_instances(state=TaskInstanceState.FAILED)]
        status = 'error' if failed_tasks else 'success'

        # TBD: use of database-neutral statement
        template = get_predefined_template('session_close.sql')
        sql = template.render(
            session_id=self.session.session_id,
            status=status,
            finished=datetime.now().isoformat()
        )
        hook: DbApiHook = DbApiHook.get_hook(self.session.session_conn_id)
        hook.run(sql)
        self.log.info(f'Session {self.session.session_id} closed with status {status}')

        if status == 'error':
            raise AirflowException('Setting drives to idle')

def open_session(
        source_conn_id: str, 
        destination_conn_id: str, 
        session_conn_id: str | None = None, 
        dag: DAG | None = None,
        **kwargs) -> XComArg:
    """ Opens a new ETL session.

    ETL session is a logical group of data transfers united by single identifier (`session_id`).
    Sessions store information about data transfer source, target, data loading period
    and completion state thus providing all necessary information about the data flow.

    Call to the `open_session` should be the 1st call in a data transfer DAG. It will
    create a new session by adding a record to `sessions` table and save a `ETLSession` instance 
    to XCom under "session" key. When using `ETLSession` class as a context manager,
    the call is performed implicitly.

    The session object could be retrieved from XCom and used within the queries.
        
    Note that ETL session object has `period_start` and `period_end` fields, which
    are calculated either automatically or could be specified manually by adding
    `{"period": "[<period_start>, <period_end>]"}` parameter when DAG is started. 
    These fields can be used in data extraction queries to limit dataset like this:

    ```sql
    select * from data_table 
    where some_date 
        between '{{ ti.xcom_pull(key="session").period_start }}'::timestamp
        and '{{ ti.xcom_pull(key="session").period_end }}'::timestamp
    ```

    Technically, sessions are stored in a `public.sessions` table. Table DDL (for Postgres):

    ```sql
    create table public.sessions(
        session_id serial not null primary key,
        source text not null,
        target text not null,
        period timestamptz[2] not null,
        run_id text,
        started timestamptz not null,
        finished timestamptz,
        status varchar(10) not null 
            check (status in ('running', 'success', 'error'))
    );
    ```

    Every table where the data is saved should have an extra `session_id` field 
    referencing the `sessions` table. For example:

        create table public.test_table(
            session_id int not null references public.sessions(session_id),
            id int not null,
            name text not null
        );
    
    This allows to easily identify when particular record was loaded or 
    clean up after unsuccessfull attempts.

    Args:
        source_conn_id:   Airflow connection where source data resides
        destination_conn_id:   Airflow connection to transfer data to
        session_conn_id:   ID of connection, where `sessions` table is located.
            If not set, then `destination_conn_id` is used.

    Returns:
        An `XComArg` placeholder object indicating the session was created 
        - due to Airflow's architecture actual session object could not be accessed 
        at this point, but it will automatically be converted to real one
        upon passing in to the TaskFlow's task function (see examples).

    Examples:
        Create a DAG which opens a session, outputs info to log and closes it:

        >>> with DAG(...) as dag, ETLSession('source', 'target') as session:
        >>>     @dag.task
        >>>     def print_session(session: ETLSession):
        >>>         print(session)
        >>>     print_session(session)

        Create a DAG with `open-session -> transfer-test_table -> close_session`
        task sequence:

        >>> with DAG(...) as dag:
        >>>     session = open_session('source', 'target')
        >>>     transfer_table('test_table', session=session)
        >>>     close_session(session)

        Connections named `source` and `target` must be defined pointing to corresponding
        databases. Table `public.sessions` must exists in the target database.

        Table `test_table` must have the same structure in both databases, except
        that `session_id` field must be added to the target table as 1st column.

        See `astro_extras.operators.table.transfer_table` function for details on transfer operation.
    """
    
    assert (dag := dag or DagContext.get_current_dag())
    return XComArg(OpenSessionOperator(
        dag=dag,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session_conn_id=session_conn_id,
        **kwargs))

def close_session(
        session: ETLSession | XComArg, 
        upstream: Any | None = None,
        dag: DAG | None = None,
        **kwargs) -> XComArg:
    """ Closes the ETL session.

    Updates `public.sessions` table for currently running session
    saving completion time and state. If all tasks in a DAG run were successfull,
    session's state will be set to `success`, otherwise - to `error`. Note that 
    session closing task will also fail in later case in order to be
    automatically included into further retrying attempts.
    
    This function should be the last call in a DAG, and it will automatically try to link up 
    to the end of the task chain.
    
    If a `ETLSession` class is used as a context manager, `close_session` will be called
    implicitly.

    Args:
        session:    An object returned by `open_session`
        upstream_task:  A DAGs task which `close_session` task
            would be linked to. If not set, a last DAG's task will be used.

    Returns:
        An `XComArg` object indicating the session was closed

    Examples:
        See `open_session` for examples
    """
    
    assert (dag := dag or DagContext.get_current_dag())
    if not upstream:
        if not len(dag.tasks):
            raise ValueError('close_session must not be the first DAG operator')
        upstream = dag.tasks[-1]
        if upstream.task_group and not upstream.task_group.is_root:
            upstream = upstream.task_group

    op = CloseSessionOperator(dag=dag, session=session, **kwargs)
    op.set_upstream(upstream)
    return XComArg(op)

def get_current_session(context: Context | None = None) -> ETLSession:
    """ Retrieves current session from XCom.
    
    Args:
        context:    DAG execution context (optional)

    Returns:
        Current ETL session as `ETLSession` class instance
    """
    context = context or get_current_context()
    return context['ti'].xcom_pull(key='session')

def ensure_session(session: ETLSession | XComArg | None, 
                   context: Context | None = None) -> ETLSession:
    """ Returns current session. If a placeholder object returned by `open_session` is passed in,
    retrieves actual session from XCom. 
    
    Args:
        session: Either the `ETLSession` instance or a placeholder returned by `open_session` call
        context:    DAG execution context (optional)

    Returns:
        Current ETL session as `ETLSession` class instance
    """
    match session:
        case None:
            return None
        case ETLSession():
            return session
        case XComArg():
            return get_current_session(context)
        case _:
            raise TypeError(f'Either ETLSession or XComArg expected, {session.__class__.__name__} found')

_TS_REGX = r'\d{4}-([0]\d|1[0-2])-([0-2]\d|3[01])(T\d{2}:\d{2}:\d{2})?'
_FULL_REGX = r'\[\d{4}-([0]\d|1[0-2])-([0-2]\d|3[01])(T\d{2}:\d{2}:\d{2})?,\s*\d{4}-([0]\d|1[0-2])-([0-2]\d|3[01])(T\d{2}:\d{2}:\d{2})?]'

def get_session_period(context: Context | None = None) -> Tuple[str, str]:
    """ Calculates ETL session loading period.
    
    This function is used when a new session is created. It recognizes a "period"
    DAG run parameter, which must be provided as two valid dates or datetimes defining
    lower and upper bound of loading period. If a date-only upper bound is used,
    it will be increased to hold entire day (see examples).
    
    If this option was not specified, loading period will be set as interval of
    `[data_interval_start, data_interval_end]` Airflow variables 
    (see https://docs.astronomer.io/learn/scheduling-in-airflow for details).
    However, these dates will be converted to Airflow default timezone 
    (as they are defined in UTC).

    Args:
        context:    DAG execution context (optional)

    Returns:
        Tuple of two datetimes indicating lower- and upper-bound of loading period,
        converted to ISO-formatted strings (see `datetime.isoformat()`).

    Examples:
        Examples of DAG run configuration options and their conversion:

        ```
        {"period": "[2023-05-01, 2023-05-31]"}
            -> ["2023-05-01T00:00:00", "2023-06-01T00:00:00"]
        {"period": "[2023-05-01, 2023-05-01]"} 
            -> ["2023-05-01T00:00:00", "2023-05-02T00:00:00"]
        {"period": "[2023-05-01T12:00:00, 2023-05-01T14:00:00]"} 
            -> ["2023-05-01T10:00:00", "2023-05-01T14:00:00"]
        ```

    """
    context = context or get_current_context()

    if (period_str := context['dag_run'].conf.get('period')):
        if not re.match(_FULL_REGX, period_str):
            raise AirflowFailException('Period must be specified as {"period": "[<from>,<to>]"}, got %s instead', period_str)
        
        period = [isoparse(x.group(0)) for x in re.finditer(_TS_REGX, period_str)]
        if len(period) < 2:
            raise AirflowFailException('Invalid period: two valid dates in ISO format must be specified, got %s instead', period_str)

        # If no time part is provided in the upper period bound,
        # consider this to be date-only and add 1 day to align to days'end
        # For example, specifying period = ["2023-05-01", "2023-05-01"] will be converted
        # to ["2023-05-01 00:00:00", "2023-05-02 00:00:00"]
        if period[1].hour == 0 and period[1].minute == 0 and period[1].second == 0:
            period[1] += timedelta(days=1)

        if period[1] <= period[0]:
            raise AirflowFailException('Upper period bound must be greater than lower bound')
    else:
        period = [context['data_interval_start'], context['data_interval_end']]
        period = [datetime_to_tz(x, TIMEZONE) for x in period]

    return tuple([x.isoformat() for x in period])



# Astro SDK Extras project
# (c) kol, 2023-2024

""" Table operations """

import logging
import warnings
import numpy as np
import pandas as pd
from sqlalchemy import text, Table as SqlaTable, Integer, BigInteger, SmallInteger
from sqlalchemy.engine.base import Connection as SqlaConnection

from airflow.models.dag import DagContext
from airflow.utils.context import Context
from airflow.models.xcom_arg import XComArg
from airflow.utils.task_group import TaskGroup
from airflow.exceptions import AirflowFailException
from airflow.operators.generic_transfer import GenericTransfer

from astro import sql as aql
from astro.databases import create_database
from astro.sql.table import BaseTable, Table
from astro.databases.base import BaseDatabase
from astro.airflow.datasets import kwargs_with_datasets

from .session import ETLSession, ensure_session
from ..utils.utils import ensure_table, schedule_ops, is_same_database_uri, adjust_table_name_case
from ..utils.template import get_template, get_template_file, get_predefined_template
from ..utils.data_compare import compare_datasets, compare_timed_dict
from ..utils.postgres_sql import postgres_merge_tables

from typing import Iterable, Type

DEFAULT_CHUNK_SIZE: int = 10000
""" Chunk size for `pd.to_sql()` calls. Set to value lesser than `DEFAULT_CHUNK_SIZE` in Astro SDK to avoid memory overload """

class TableTransfer(GenericTransfer):
    """
    Table transfer operator to be used within `transfer_table` function.
    Implements 'bulk' data transfer without any extra conditioning.
    """

    template_fields = ("sql", "preoperator", "source_table", "destination_table")
    template_ext = (".sql", ".hql" )
    template_fields_renderers = {"sql": "sql", "preoperator": "sql"}
    ui_color = "#b4f07c"

    # Public attributes
    source_db: BaseDatabase
    """ Source database object """

    dest_db: BaseDatabase
    """ Destination database object """

    source: BaseTable
    """ Source table object """

    source_table: str
    """ Source table fully-qualified name """

    destination: BaseTable
    """ Destination table object """

    destination_table: str
    """ Destination table fully-qualified name """

    session: ETLSession | XComArg | None
    """ ETL Session. Use `ensure_session` to cast to proper session type """

    in_memory_transfer: bool
    """ Flag to use in-memory transfer (othewise it would be Airflow's cursor-based, which is several times slower) """

    chunk_size: int
    """ Chunk size for `pd.to_sql() calls """

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        in_memory_transfer: bool = False,
        proper_case: bool = False,
        chunk_size: int | None = DEFAULT_CHUNK_SIZE,
        **kwargs,
    ) -> None:

        # source and target databases
        self.source_db = create_database(source_table.conn_id)
        self.dest_db = create_database(destination_table.conn_id)

        # source and target AstroSDK tables
        if not proper_case:
            # adjust table name case to support DB requirements
            self.source = source_table
            self.destination = destination_table
        else:
            # adjust table name case to support DB requirements
            self.source = adjust_table_name_case(source_table, self.source_db)
            self.destination = adjust_table_name_case(destination_table, self.dest_db)

        # source and target fully-qualified table names
        # self.destination_table will be set by `super().__init__()`
        self.source_table = self.source_db.get_table_qualified_name(self.source)

        # task_id would be 'transfer-<schema>_<table>'
        task_id = kwargs.pop('task_id', f'transfer-{self.source_table.replace(".", "_")}')

        # sql is either passed in by the caller or rendered from template
        sql = kwargs.pop('sql', self._get_sql(self.source, self.source_db, session))

        super().__init__(task_id=task_id,
                         sql=sql,
                         destination_table=self.dest_db.get_table_qualified_name(self.destination),
                         source_conn_id=self.source.conn_id,
                         destination_conn_id=self.destination.conn_id,
                         **kwargs)

        self.session = session
        self.in_memory_transfer = in_memory_transfer
        self.chunk_size = chunk_size

        # flag to prevent multiple _pre_execute() calls
        self._pre_execute_called: bool = False

        # statistics for lineage
        self._row_count: int = 0

    def _get_sql(self, table: BaseTable, db: BaseDatabase, session: ETLSession | None = None, suffix: str | None = None) -> str:
        """ Internal - get a sql statement or template for given table """

        # Check whether a template SQL exists for given table under dags\templates\<dag_id>
        # Actual query will be loaded by Airflow templating itself
        full_name = db.get_table_qualified_name(table) + (suffix or '')
        if (sql_file := get_template_file(full_name, '.sql')) or (sql_file := get_template_file(table.name, '.sql')) :
            self.log.info(f'Using template file {sql_file}')
            return sql_file

        # Nope, load an SQL from package resources substituting template fields manually
        # SQL file names are fixed according to whether we do run under ETL session or not
        template_name = 'table_transfer_nosess.sql' if not session else 'table_transfer_sess.sql'
        template = get_predefined_template(template_name)
        return template.render(source_table=full_name)

    def _pre_execute(self, context: Context):
        """ Internal - run before execution """

        if self._pre_execute_called:
            return

        # Set debug level logging if requested upon DAG start
        # Config is: {..., "debug": true}
        assert 'dag_run' in context
        if context['dag_run'].conf.get('debug'):
            self.log.setLevel(logging.DEBUG)
            logging.getLogger('airflow.task').setLevel(logging.DEBUG)
            self.log.debug('Log level set to DEBUG')

        # Transform XComArg session into ETLSession
        self.session = ensure_session(self.session, context)
        if self.session:
            if not self.source_conn_id:
                self.source_conn_id = self.session.source_conn_id
            if not self.destination_conn_id:
                self.destination_conn_id = self.session.destination_conn_id

        if not self.source_conn_id:
            raise AirflowFailException('source connection not specified')
        if not self.destination_conn_id:
            raise AirflowFailException('destination connection not specified')
        if self.source_conn_id == self.destination_conn_id and self.source_table == self.destination_table:
            raise AirflowFailException('Source and destination must not be the same')
        
        # Load source and target table definitions
        # consider adding a flag to skip one or both - usable for slow connections
        src_meta = self.source.sqlalchemy_metadata
        src_meta.reflect(bind=self.source_db.connection, only=[self.source.name])
        if (src_sqla_table := src_meta.tables.get(self.source_table)) is None:
            raise AirflowFailException(f'Table {self.source_table} does not exist on {self.source_conn_id}')
        for c in src_sqla_table.columns:
            self.source.columns.append(c)
        self.log.info(f'Table {self.source_table} columns: {",".join([f"{c.name} {c.type}" for c in self.source.columns])}')

        dest_meta = self.destination.sqlalchemy_metadata
        dest_meta.reflect(bind=self.dest_db.connection, only=[self.destination.name])
        if (dest_sqla_table := dest_meta.tables.get(self.destination_table)) is None:
            raise AirflowFailException(f'Table {self.destination_table} does not exist on {self.destination_conn_id}')
        for c in dest_sqla_table.columns:
            self.destination.columns.append(c)
        self.log.info(f'Table {self.destination_table} columns: {",".join([f"{c.name} {c.type}" for c in self.destination.columns])}')

        self._pre_execute_called = True

    def _adjust_dtypes(self, data: pd.DataFrame, table: BaseTable) -> pd.DataFrame:
        """ Checks whether `data` column dtypes match given `table` ones and reset them to proper type.
        This could happen if, for example, source has nulls in `int` columns which would have `float` type in dataframe.
        """

        for col in table.columns:
            if col.name not in data.columns:
                self.log.warning(f'Column {col.name} of table {table.name} was not found in dataset')
                continue

            match col.type:
                case Integer() | SmallInteger() | BigInteger() if data.dtypes[col.name] == 'float':
                    self.log.info(f'Correcting column {col.name} dtype from `float` to `int`')
                    data[col.name] = data[col.name].astype('Int64')

                case _:
                    pass

        return data

    def execute(self, context: Context):
        """ Execute operator """
        self._pre_execute(context)
        if not self.in_memory_transfer:
            return super().execute(context)
        
        # upload source data to memory (might need to use chunks)
        self.log.info('Using in-memory transfer')
        data = pd.read_sql(self.sql, self.source_db.connection)
        if data is None or data.empty:
            self.log.info('No data to transfer')
            return

        # Adjust dtypes
        data = self._adjust_dtypes(data, self.destination)

        # save to target table
        self._row_count = len(data)
        self.log.info(f'{self._row_count} records to be transferred')
        self.dest_db.load_pandas_dataframe_to_table(
            data, 
            self.destination, 
            if_exists='append',
            chunk_size=self.chunk_size)

    def get_openlineage_facets_on_complete(self, task_instance):  # skipcq: PYL-W0613
        """
        Collect the input, output, job and run facets for DataframeOperator
        """
        from airflow.providers.openlineage.extractors.base import OperatorLineage
        from openlineage.client.run import Dataset
        from openlineage.client.facet import (
            BaseFacet,
            DataSourceDatasetFacet,
            OutputStatisticsOutputDatasetFacet,
            SchemaDatasetFacet,
            SchemaField,
            SqlJobFacet,
        )

        input_dataset: list[Dataset] = []
        if self.source and self.source.openlineage_emit_temp_table_event():
            input_dataset = [
                Dataset(
                    namespace=self.source.openlineage_dataset_namespace(),
                    name=self.source.openlineage_dataset_name(),
                    facets={
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaField(
                                    name=f'schema: {self.source.metadata.schema or self.source_db.default_metadata.schema}',
                                    type='string',
                                ),
                                SchemaField(
                                    name=f'database: {self.source.metadata.database or self.source_db.default_metadata.database}',
                                    type='string',
                                ),
                                *[SchemaField(name=col.name, type=str(col.type)) for col in self.source.columns],
                            ]
                        ),
                        "dataSource": DataSourceDatasetFacet(
                            name=self.source_table, uri=self.source.openlineage_dataset_uri()
                        ),
                    },
                ),
            ]

        output_dataset: list[Dataset] = []
        if self.destination and self.destination.openlineage_emit_temp_table_event():  # pragma: no cover
            output_dataset = [
                Dataset(
                    namespace=self.destination.openlineage_dataset_namespace(),
                    name=self.destination.openlineage_dataset_name(),
                    facets={
                        "schema": SchemaDatasetFacet(
                            fields=[
                                SchemaField(
                                    name=f'schema: {self.destination.metadata.schema or self.dest_db.default_metadata.schema}',
                                    type='string',
                                ),
                                SchemaField(
                                    name=f'database: {self.destination.metadata.database or self.dest_db.default_metadata.database}',
                                    type='string',
                                ),
                                *[SchemaField(name=col.name, type=str(col.type)) for col in self.destination.columns],
                            ]
                        ),
                        "dataSource": DataSourceDatasetFacet(
                            name=self.destination_table, uri=self.destination.openlineage_dataset_uri()
                        ),
                        "outputStatistics": OutputStatisticsOutputDatasetFacet(
                            rowCount=self._row_count,
                        ),
                    },
                ),
            ]

        run_facets: dict[str, BaseFacet] = {}
        job_facets: dict[str, BaseFacet] = {
            "sql": SqlJobFacet(query=str(self.sql))
        }
        return OperatorLineage(
            inputs=input_dataset, outputs=output_dataset, run_facets=run_facets, job_facets=job_facets
        )

class ChangedTableTransfer(TableTransfer):
    """
    Table transfer operator to be used within `transfer_changed_table` function.
    Compares source and target data and transfers only if any changes detected.
    Requiures `xxx_a` view to exist on target.
    """

    template_fields = ("sql", "preoperator", "source_table", "destination_table", "destination_sql")
    template_fields_renderers = {"sql": "sql", "preoperator": "sql", "destination_sql": "sql"}
    ui_color = "#95f07c"

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        **kwargs,
    ) -> None:

        dest_db = create_database(destination_table.conn_id, destination_table)
        dest_sql = kwargs.pop('destination_sql', self._get_sql(destination_table, dest_db, suffix='_a'))

        super().__init__(source_table=source_table,
                         destination_table=destination_table,
                         session=session,
                         **kwargs)

        self.destination_sql: str = dest_sql

    def _compare_datasets(self, src_conn: SqlaConnection, dest_conn: SqlaConnection, stop_on_first_diff: bool, logger: logging.Logger | None = None):
        """ Internal - compare source and target dictionaries """

        logger = logger or self.log

        logger.info(f'Executing: {self.sql}')
        df_src = pd.read_sql(self.sql, src_conn)
        logger.info(f'Executing: {self.destination_sql}')
        df_trg  = pd.read_sql(self.destination_sql, dest_conn)

        return compare_datasets(df_src, df_trg, stop_on_first_diff=stop_on_first_diff, logger=logger)

    def execute(self, context: Context):
        """ Execute operator """
        self._pre_execute(context)
        if not self._compare_datasets(self.source_db.connection, self.dest_db.connection, stop_on_first_diff=True):
            return super().execute(context)

class OdsTableTransfer(ChangedTableTransfer):
    """
    Table transfer operator to be used within `transfer_ods_table` function
    to transfer data from source to ODS-style target.
    
    Requiures `xxx_a` view to exist on target. Target table must have
    `_modified' and `_deleted` attributes of `timestamp` or `timestamptz` type.
    """
    ui_color = "#78f07c"

    def _save_data(self, data: pd.DataFrame, conn: SqlaConnection, modified: pd.Timestamp, deleted: pd.Timestamp | None, category: str) -> None:
        if data is not None and not data.empty:
            if self.session is not None:
                data.drop(columns=set(['session_id', '_modified', '_deleted']) & set(data.columns), inplace=True)
                data.insert(0, '_deleted', deleted)
                data.insert(0, '_modified', modified)
                data.insert(0, 'session_id', self.session.session_id)
            else:
                data['_modified'] = modified
                data['_deleted'] = deleted

            self.log.info(f'Saving {data.shape[0]} {category} records to {self.destination_table} (chunk size is {self.chunk_size})')
            data.to_sql(
                self.destination.name,
                con=conn,
                schema=self.destination.metadata.schema if self.destination.metadata else None,
                if_exists='append',
                method='multi',
                chunksize=self.chunk_size,
                index=False,
            )
            self._row_count += len(data)

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks
        self._pre_execute(context)

        # Verify target table structure
        n_col = 0
        if self.session is not None:
            if self.destination.columns[n_col].name.lower() != 'session_id':
                raise AirflowFailException(f'Invalid ODS table {self.destination_table} structure: `session_id` missing or at improper place')
            n_col += 1
        if self.destination.columns[n_col].name.lower() != '_modified':
            raise AirflowFailException(f'Invalid ODS table {self.destination_table} structure: `_modified` column missing or at improper place')
        if self.destination.columns[n_col+1].name.lower() != '_deleted':
            raise AirflowFailException(f'Invalid ODS table {self.destination_table} structure: `_deleted` column missing or at improper place')

        # compare datasets and save delta frames to target (new/modified/deleted)
        self._row_count = 0
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn:
            dfn, dfm, dfd = self._compare_datasets(src_conn, dest_conn, stop_on_first_diff=False)
            with dest_conn.begin():
                self._save_data(dfn, dest_conn, pd.Timestamp.utcnow(), pd.NaT, 'new')
                self._save_data(dfm, dest_conn, pd.Timestamp.utcnow(), pd.NaT, 'modified')
                self._save_data(dfd, dest_conn, pd.Timestamp.utcnow(), pd.Timestamp.utcnow(), 'deleted')

class ActualsTableTransfer(TableTransfer):
    """
    Table transfer operator to be used within `transfer_actuals_table` function.
    Source and target table must have `_deleted` attribute of `timestamp` or `timestamptz` type.
    """

    ui_color = "#5af07d"

    def __init__(
        self,
        *,
        source_table: BaseTable,
        destination_table: BaseTable,
        session: ETLSession | None = None,
        as_ods: bool = False,
        keep_temp_table: bool = False,
        replace_data: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(source_table=source_table, destination_table=destination_table, session=session, **kwargs)
        self.as_ods = as_ods
        self.keep_temp_table = keep_temp_table
        self.replace_data = replace_data

    def _get_sql(self, table: BaseTable, db: BaseDatabase, session: ETLSession | None = None, suffix: str | None = None) -> str:
        """ Internal - get a sql statement or template for given table """

        # Get template SQL from package resources
        # The template defines `select` query which takes data from source limited 
        # to one's loaded with successfull sessions with the same loading period as this operator has
        full_name = db.get_table_qualified_name(table)
        template = get_predefined_template('table_actuals_select.sql')
        return template.render(source_table=full_name)

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks
        self._pre_execute(context)

        # Verify structure
        if self.as_ods:
            # Check source and target meet the ODS requirements
            if not [c for c in self.source.columns if c.name == '_deleted']:
                raise AirflowFailException(f'Source table {self.source_table} does not have `_deleted` column required for ODS-style transfer')
            if not [c for c in self.destination.columns if c.name == '_deleted']:
                raise AirflowFailException(f'Target table {self.destination_table} does not have `_deleted` column required for ODS-style transfer')

        # Build a source-to-target column mapping
        col_map = {}
        id_cols = []
        for src_col in self.source.columns:
            dest_cols = [c for c in self.destination.columns if c.name.lower() == src_col.name.lower()]
            if not dest_cols:
                self.log.warning(f'Column {src_col.name} was not found in {self.destination_table} table, skipping')
            else:
                # If a column is PK, store it to use in `on conflict` statement part
                col_map[src_col.name] = dest_cols[0].name
                if dest_cols[0].primary_key:
                    id_cols.append(dest_cols[0].name)

        # Check there are any ID columns on target
        if not id_cols:
            raise AirflowFailException(f'Could not detect primary key on {self.destination_table}')

        # Replace `t.id` with ids and `t.*` with selection columns list and proper session_id
        id_cols_str = ", ".join(id_cols)
        all_cols_str = f'{self.session.session_id} as session_id, ' + ", ".join([col for col in col_map if col != 'session_id'])
        sql = self.sql.replace('t.id', id_cols_str).replace('t.*', all_cols_str)
        self.log.info(f'Source extraction SQL:\n{sql}')

        # Check whether the hooks point to the same database
        same_db = is_same_database_uri(self.source_db.hook.get_uri(), self.dest_db.hook.get_uri())

        # Update/merge
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn:
            if same_db:
                # Source and destination tables are in the same database, 
                self.log.info(f'Source and destination tables are in the same database')
                with dest_conn.begin():
                    # If `replace_data` is set, purge target table
                    if self.replace_data:
                        dest_conn.execute(text(f'delete from {self.destination_table}'))

                    # Merge into target table directly using source sql
                    self._row_count = postgres_merge_tables(
                        conn=dest_conn,
                        source_table=None,
                        target_table=self.destination,
                        source_to_target_columns_map=col_map,
                        target_conflict_columns=id_cols,
                        source_sql=sql,
                        if_conflicts='update'
                    )
                    self.log.info(f'{self._row_count} records transferred')

            else:
                # Source and destination tables are in different databases
                temp_table = Table(metadata=self.destination.metadata, temp=True)
                self.log.info(f'Source and destination tables are in different databases, transferring via temporary table')

                # Load data from source table
                data = pd.read_sql(sql, src_conn)
                if data is None or data.empty:
                    self.log.info('No data to transfer')
                    return
                
                if self.session:
                    data['session_id'] = self.session.session_id
                    col_map['session_id'] = 'session_id'

                # Temp table has to be created 1st, otherwise it might be column types mismatch
                temp_sqla_table = SqlaTable(temp_table.name, temp_table.sqlalchemy_metadata, *([c.copy() for c in self.source.columns]))
                self.destination.sqlalchemy_metadata.create_all(bind=dest_conn, tables=[temp_sqla_table], checkfirst=False)

                try:
                    # Save data to temporary table
                    data.to_sql(
                        temp_table.name,
                        con=dest_conn,
                        schema=temp_table.metadata.schema if temp_table.metadata else None,
                        if_exists='append',
                        method='multi',
                        chunksize=self.chunk_size,
                        index=False,
                    )

                    with dest_conn.begin():
                        # If `replace_data` is set, purge target table
                        if self.replace_data:
                            dest_conn.execute(text(f'delete from {self.destination_table}'))

                        # Merge temporary and target tables
                        self.log.info(f'Merging from {self.dest_db.get_table_qualified_name(temp_table)} to {self.destination_table}')
                        self._row_count = postgres_merge_tables(
                            conn=dest_conn,
                            source_table=temp_table,
                            target_table=self.destination,
                            source_to_target_columns_map=col_map,
                            target_conflict_columns=id_cols,
                            if_conflicts='update'
                        )
                        self.log.info(f'{self._row_count} records transferred')

                finally:
                    if not self.keep_temp_table:
                        self.dest_db.drop_table(temp_table)

class TimedTableTransfer(ChangedTableTransfer):
    """
    Table transfer operator to be used within `update_timed_dict` function.
    Implements timed dictionary update behaviour.
    Requiures `xxx_a` view to exist both on source and target.
    """
    ui_color = "#3cf07e"

    def execute(self, context: Context):
        """ Execute operator """

        # All the checks
        self._pre_execute(context)

        # Get source and target data
        with self.source_db.connection as src_conn, self.dest_db.connection as dest_conn, dest_conn.begin():
            self.log.info(f'Executing: {self.sql}')
            df_src = pd.read_sql(self.sql, src_conn)
            self.log.info(f'Executing: {self.destination_sql}')
            df_trg = pd.read_sql(self.destination_sql, dest_conn)

            # Get the deltas
            df_opening, df_closing = compare_timed_dict(df_src, df_trg)
            if (df_opening is None or df_opening.empty) and (df_closing is None or df_closing.empty):
                self.log.info('No changes detected, nothing to do')
                return

            # Update records to be closed
            if df_closing is not None and not df_closing.empty:
                update_sql_template = get_predefined_template('table_timed_update.sql')
                update_sql = update_sql_template.render(destination_table=self.destination_table)
                update_params = df_closing.to_dict(orient='list')
                dest_conn.execute(text(update_sql).execution_options(autocommit=True), update_params)

            # Insert new records
            if df_opening is not None and not df_opening.empty:
                df_opening.to_sql(
                    self.destination.name,
                    con=dest_conn,
                    schema=self.destination.metadata.schema if self.destination.metadata else None,
                    if_exists='append',
                    method='multi',
                    chunksize=self.chunk_size,
                    index=False,
                )


class CompareTableOperator(ChangedTableTransfer):
    """
    Table comparsion operator to be used within `compare_table` function.
    Compares source and target data and prints results to log.
    """
    ui_color = "#64bf62"

    def execute(self, context: Context):
        """ Execute operator """
        logger = logging.getLogger(f'compare_tables_logger')
        logger.setLevel(logging.DEBUG)
        if not self._compare_datasets(self.source_db.connection, self.dest_db.connection, stop_on_first_diff=True, logger=logger):
            raise AirflowFailException(f'Differences detected')

def load_table(
        table: str | BaseTable,
        conn_id: str | None = None,
        session: ETLSession | None = None,
        sql: str | None = None) -> XComArg:
    """ Loads table into memory.

    This is a wrapper over Astro-SDK `run_raw_sql` to
    load data from given database table into XCom and make it available for 
    further processing.

    SQL templating is supported, e.g. if a template for given table was found, it
    will be executed to get the data (see `astro_extras.utils.template.get_template_file`).

    Please note that in order to operate even on modest volumes of data,
    intermediate XCom storage might be required. Easiest way to set it up is to use
    local Parquet file storage by setting
    `AIRFLOW__ASTRO_SDK__XCOM_STORAGE_CONN_ID=local` environment
    variable. However, this will add extra serialization/deserialization
    operation to every task thus increasing overall DAG execution time.

    See https://astro-sdk-python.readthedocs.io/en/1.2.0/guides/xcom_backend.html
    for details.

    Args:
        table:  Either a table name or Astro-SDK `Table` object to load data from
        conn_id:    Airflow connection ID to underlying database. If not specified,
            and `Table` object is passed it, its `conn_id` attribute will be used.
        session:    `astro_extras.operators.session.ETLSession` object. 
            Used only to link up to the `open_SESSIONion` operator.
        sql:    Custom SQL to load data, used only if no SQL template found.
            If neither SQL nor template is given, all table data will be loaded.

    Results:
        `XComArg` object suitable for further manipulations with Astro-SDK functions

    Examples:
        >>> @aql.dataframe
        >>> def modify_data(data: pd.DataFrame):
        >>>     data['some_column'] = 'new_value'
        >>>     return data
        >>> data = load_table('test_table', conn_id='source_db')
        >>> modified_data = modify_data(data)
        >>> save_table(modified_data, conn_id='target_db')
    """

    if not isinstance(table, BaseTable):
        dag = DagContext.get_current_dag()
        @aql.run_raw_sql(handler=lambda result: result.fetchall(),
                        conn_id=conn_id,
                        task_id=f'load-{table}',
                        results_format='pandas_dataframe')
        def _load_table_by_name(table: str, session: ETLSession | None):
            sql_file = get_template_file(table, '.sql', dag=dag)
            return sql or sql_file or f'select * from {table}'

        return _load_table_by_name(table, session)
    else:
        dag = DagContext.get_current_dag()
        @aql.run_raw_sql(handler=lambda result: result.fetchall(),
                        conn_id=conn_id,
                        task_id=f'load-{table.name}',
                        results_format='pandas_dataframe')
        def _load_table(table: BaseTable, session: ETLSession | None):
            sql_file = get_template_file(table.name, '.sql', dag=dag)
            return sql or sql_file or '''select * from {{table}}'''

        return _load_table(table, session)

def save_table(
        data: XComArg,
        table: str | BaseTable,
        conn_id: str | None = None,
        session: ETLSession | None = None,
        fail_if_not_exist: bool | None = True) -> XComArg:
    """ Saves a table into database """

    table = ensure_table(table, conn_id)
    task_id = f'save-{table.name}'
    conn_id = conn_id or table.conn_id

    @aql.dataframe(if_exists='append', conn_id=conn_id, task_id=task_id)
    def _save_data(data: pd.DataFrame, session: ETLSession):
        if fail_if_not_exist:
            db = create_database(conn_id, table)
            if not db.table_exists(table):
                raise AirflowFailException(f'Table {table.name} was not found under {conn_id} connection')
        session = ensure_session(session)
        if session and 'session_id' not in data.columns:
            data.insert(0, 'session_id', session.session_id)
        return data

    return _save_data(data, session, output_table=table)

def declare_tables(
        table_names: Iterable[str],
        conn_id: str | None = None,
        schema: str | None = None,
        database: str | None = None,
) -> Iterable[BaseTable]:
    """ Convert list of string table names to list of `Table` objects """

    return [ensure_table(t, conn_id, schema, database) for t in table_names]

def _do_transfer_table(
        op_cls: Type[GenericTransfer],
        source: str | BaseTable,
        target: str | BaseTable | None,
        source_conn_id: str | None,
        destination_conn_id: str | None,
        session: XComArg | ETLSession | None,
        **kwargs) -> XComArg:
    """ Internal - table transfer implementation """

    source_table = ensure_table(source, source_conn_id)
    dest_table = ensure_table(target or source_table, destination_conn_id)

    op = op_cls(
        source_table=source_table,
        destination_table=dest_table,
        session=session,
        **kwargs_with_datasets(kwargs=kwargs, 
                               input_datasets=source_table, 
                               output_datasets=dest_table)
    )
    return XComArg(op)

def _do_transfer_tables(
        op_cls: Type[GenericTransfer],
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None,
        source_conn_id: str | None,
        destination_conn_id: str | None,
        group_id: str | None,
        num_parallel: int,
        session: XComArg | ETLSession | None,
        **kwargs) -> TaskGroup:
                 
    """ Internal - transfer multiple tables """

    if target_tables and len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    target_tables = target_tables or source_tables

    with TaskGroup(group_id, add_suffix_on_collision=True, prefix_group_id=False) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id) or source_table
            op = op_cls(
                source_table=source_table,
                destination_table=dest_table,
                session=session,
                **kwargs_with_datasets(
                    kwargs=kwargs, 
                    input_datasets=source_table, 
                    output_datasets=dest_table))
            ops_list.append(op)

        schedule_ops(ops_list, num_parallel)

    return tg

def transfer_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        changes_only: bool | None = None,
        **kwargs) -> XComArg:
    
    """ Cross-database data transfer.

    This function implements cross-database geterogenous data transfer.

    It reads data from source table into memory and then sequentaly inserts 
    each record into the target table. Fields order in the source and target tables 
    must be identical and field types must be compatible, or transfer will fail or produce 
    undesirable results.

    To limit data selection or customize fields, a SQL template could be 
    created for the source table (see `astro_extras.utils.template.get_template_file`).
    The template must ensure fields order and type compatibility with the target table.
    If transfer is running under `astro_extras.operators.session.ETLSession` context, 
    a `session_id` field must also be manually added at proper place.

    For example, if these tables are to participate in transfer:

    ``` sql
    create table source_data ( a int, b text );
    create table target_data ( session_id int, b text, a int );
    ```

    then, this SQL template might be created as `source_data.sql` file:

    ``` sql
    select {{ti.xcom_pull(key="session").session_id}} as session_id, b, a 
    from source_data;
    ```

    Args:
        source: Either a table name or a `Table` object which would be a data source.
            If a string name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If a SQL template exists for this table name, it will be executed,
            otherwise all table data will be selected.

        target: Either a table name or a `Table` object where data will be saved into.
            If a name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If omitted, `source` argument value is used (this makes sense only
            with string table name and different connections).

        source_conn_id: Source database Airflow connection.
            Used only with string source table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.source_conn_id` will be used.

        destination_conn_id: Destination database Airflow connection.
            Used only with string target table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.destination_conn_id` will be used.

        session:    `ETLSession` object. If set and no SQL template is defined,
            a `session_id` field will be automatically added to selection.

        changes_only:   If set to `True`, the operator will compare source and target
            tables and transfer data only when they are different. Target data are obtained
            by runnning `destination_sql`. By default, this query will be built using 
            `<destination_table>_a` view to get actual data.

            Deprecated since 0.1.1, use `transfer_changed_table` instead.

        kwargs:     Any parameters passed to underlying operator (e.g. `preoperator`, ...)

    Returns:
        `XComArg` object

    Examples:
        Using `Table` objects (note use of `Metadata` object to specify schemas):

        >>> with DAG(...) as dag:
        >>>     input_table = Table('table_data', conn_id='source_db', 
        >>>                          metadata=Metadata(schema='public'))
        >>>     output_table = Table('table_data', conn_id='target_db', 
        >>>                          metadata=Metadata(schema='stage'))
        >>>     transfer_table(input_table, output_table)

        Using string table name:

        >>> with DAG(...) as dag, ETLSession('source_db', 'target_db') as sess:
        >>>     transfer_table('public.table_data', session=sess)

    """
    if changes_only is not None:
        warnings.warn('`changes_only` is deprecated since 0.0.13, use `transfer_changed_table()` instead')

    op_cls = TableTransfer if not changes_only else ChangedTableTransfer
    return _do_transfer_table(
        op_cls=op_cls,
        source=source,
        target=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        **kwargs)

def transfer_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        changes_only: bool | None = None,
        **kwargs) -> TaskGroup:

    """ Transfer multiple tables.

    Creates an Airflow task group with transfer tasks for each table pair 
    from `source_tables` and `target_tables` lists.

    See `transfer_table` for more information.
    """
    if changes_only is not None:
        warnings.warn('`changes_only` is deprecated since 0.1.1, use `transfer_changed_tables()` instead')

    op_cls = TableTransfer if not changes_only else ChangedTableTransfer
    return _do_transfer_tables(
        op_cls=op_cls,
        source_tables=source_tables,
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id or 'transfer-tables',
        num_parallel=num_parallel,
        session=session,
        **kwargs
    )

def transfer_changed_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Cross-database changed data transfer.

    This function implements cross-database geterogenous data transfer using snap-shotting
    mechanism, which is then any change is detected on source, whole source data is transferred
    to target making a new snapshot there.

    In order to detect these changes, it will load both source and target data and compare
    row count, structure and each and every record.

    This mode could be used to transfer small-sized tables (dictionaries), especially if
    the table does not contain any change mark.

    The operator uses templated SQLs to retrieve data from source and target tables
    (`sql` and `destination_sql` respectively). By default, it expects that an
    "actual data view" named `<destination_table>_a` to exist on target. This view
    has to return latest target data snapshot (usually related to last successfull ETL session),
    see example DAGs for more details.

    See `transfer_table` for details on arguments.

    Returns:
        `XComArg` object

    """
    return _do_transfer_table(
        op_cls=ChangedTableTransfer,
        source=source,
        target=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        **kwargs)

def transfer_changed_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Transfer multiple changed tables.

    Creates an Airflow task group with transfer tasks for each table pair 
    from `source_tables` and `target_tables` lists.

    See `transfer_changed_table` for more information.
    """
    return _do_transfer_tables(
        op_cls=ChangedTableTransfer,
        source_tables=source_tables,
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id or 'transfer-dict',
        num_parallel=num_parallel,
        session=session,
        **kwargs
    )

def transfer_ods_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Cross-database ODS-like data transfer.

    Returns:
        `XComArg` object

    """
    return _do_transfer_table(
        op_cls=OdsTableTransfer,
        source=source,
        target=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        **kwargs)

def transfer_ods_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Transfer multiple ODS tables.

    Creates an Airflow task group with transfer tasks for each table pair 
    from `source_tables` and `target_tables` lists.

    See `transfer_ods_table` for more information.
    """
    return _do_transfer_tables(
        op_cls=OdsTableTransfer,
        source_tables=source_tables,
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id or 'transfer-ods-tables',
        num_parallel=num_parallel,
        session=session,
        **kwargs
    )

def transfer_actuals_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession = None,
        as_ods: bool = False,
        keep_temp_table: bool = False,
        replace_data: bool = False,
        **kwargs) -> XComArg:
    
    """ Transfer table from stage to actuals.

    Returns:
        `XComArg` object

    """
    assert session is not None, 'Transfer to actuals requires ETL session'
    kwargs['as_ods'] = as_ods
    kwargs['keep_temp_table'] = keep_temp_table
    kwargs['replace_data'] = replace_data

    return _do_transfer_table(
        op_cls=ActualsTableTransfer,
        source=source,
        target=target,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        session=session,
        **kwargs)

def transfer_actuals_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        as_ods: bool = False,
        keep_temp_table: bool = False,
        replace_data: bool = False,
        **kwargs) -> TaskGroup:

    """ Transfer multiple tables from stage to actuals.
    """
    assert session is not None, 'Transfer to actuals requires ETL session'
    kwargs['as_ods'] = as_ods
    kwargs['keep_temp_table'] = keep_temp_table
    kwargs['replace_data'] = replace_data

    return _do_transfer_tables(
        op_cls=ActualsTableTransfer,
        source_tables=source_tables,
        target_tables=target_tables,
        source_conn_id=source_conn_id,
        destination_conn_id=destination_conn_id,
        group_id=group_id or 'transfer-actuals',
        num_parallel=num_parallel,
        session=session,
        **kwargs
    )

def compare_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        **kwargs) -> TaskGroup:

    """ Compare two tables.

    Creates a task of `CompareTableTransfer` operator to compare
    given tables, which would fail if any differences
    are detected among them.

    Args:
        source: Either a table name or a `Table` object which would be a data source.
            If a string name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If a SQL template exists for this table name, it will be executed,
            otherwise all table data will be selected.

        target: Either a table name or a `Table` object where data will be saved into.
            If a name is provided, it may contain schema definition denoted by `.`. 
            For `Table` objects, schema must be defined in `Metadata` field,
            otherwise Astro SDK might fall to use its default schema.
            If omitted, `source` argument value is used (this makes sense only
            with string table name and different connections).

        source_conn_id: Source database Airflow connection.
            Used only with string source table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.source_conn_id` will be used.

        destination_conn_id: Destination database Airflow connection.
            Used only with string target table name; for `Table` objects, `conn_id` field is used.
            If omitted and `session` argument is provided, `session.destination_conn_id` will be used.

    """
    source_table = ensure_table(source, source_conn_id)
    dest_table = ensure_table(target or source_table, destination_conn_id)

    op = CompareTableOperator(
        source_table=source_table, 
        destination_table=dest_table, 
        task_id=f'compare-{source_table.name}')
    return XComArg(op)

def compare_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        **kwargs) -> TaskGroup:

    """ Compares multiple tables.

    Creates an Airflow task group consisting of `CompareTableTransfer` operators 
    for each table pair from `source_tables` and `target_tables` lists.
    Each operator will compare source and target table and fails if any differences
    are detected among them.

    See `compare_table` for details.
    """
    if not target_tables or len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    with TaskGroup(group_id or 'compare-tables', add_suffix_on_collision=True) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id) or source_table
            op = CompareTableOperator(
                source_table=source_table, 
                destination_table=dest_table, 
                task_id=f'compare-{source_table.name}')
            ops_list.append(op)
        schedule_ops(ops_list, num_parallel)
    return tg

def update_timed_table(
        source: str | BaseTable,
        target: str | BaseTable | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> XComArg:
    
    """ Updates timed dictionary table (experimental) """

    source_table = ensure_table(source, source_conn_id)
    dest_table = ensure_table(target, destination_conn_id) or source_table
    op = TimedTableTransfer(
        source_table=source_table,
        destination_table=dest_table,
        session=session,
        **kwargs_with_datasets(kwargs=kwargs, 
                               input_datasets=source_table, 
                               output_datasets=dest_table)
    )
    return XComArg(op)

def update_timed_tables(
        source_tables: Iterable[str | Table],
        target_tables: Iterable[str | Table] | None = None,
        source_conn_id: str | None = None,
        destination_conn_id: str | None = None,
        group_id: str | None = None,
        num_parallel: int = 1,
        session: XComArg | ETLSession | None = None,
        **kwargs) -> TaskGroup:

    """ Updates multiple timed dictionary tables (experimental )"""

    if target_tables and len(target_tables) != len(source_tables):
        raise AirflowFailException(f'Source and target tables list size must be equal')

    target_tables = target_tables or source_tables

    with TaskGroup(group_id or 'transfer-tables', add_suffix_on_collision=True) as tg:
        ops_list = []
        for (source, target) in zip(source_tables, target_tables):
            source_table = ensure_table(source, source_conn_id)
            dest_table = ensure_table(target, destination_conn_id) or source_table
            op = TimedTableTransfer(
                source_table=source_table,
                destination_table=dest_table,
                session=session,
                **kwargs_with_datasets(
                    kwargs=kwargs, 
                    input_datasets=source_table, 
                    output_datasets=dest_table))
            ops_list.append(op)
        schedule_ops(ops_list, num_parallel)
    return tg




# Astro SDK Extras project
# (c) kol, 2023

""" Misc utility functions """

from urllib.parse import urlparse
from itertools import pairwise
from airflow.models import BaseOperator
from astro.sql.table import BaseTable, Table, Metadata
from astro.databases.base import BaseDatabase
from typing import Union, Tuple, Optional, List, Iterable

def split_table_name(table: str) -> Tuple[Optional[str], str]:
    """ Splits table name to schema and table name """
    if not table:
        raise ValueError('Table name not specified')
    parts = table.split('.', 1)
    return tuple(parts) if len(parts) > 1 else (None, parts[0])

def ensure_table(
        table: Union[str, BaseTable], 
        conn_id: Optional[str] = None, 
        schema: Optional[str] = None,
        database: Optional[str] = None) -> BaseTable:
    
    """ Ensure an object is a table """
    if table is None:
        return None
    
    if isinstance(table, BaseTable):
        return table
    elif isinstance(table, str):
        schema_from_name, table = split_table_name(table)
        return Table(table, conn_id=conn_id, metadata=Metadata(schema=schema_from_name or schema, database=database))
    else:
        raise TypeError(f'Either str or BaseTable expected, {table.__class__.__name__} found')

def split_iterable(it: Iterable, num_splits: int = 1) -> Iterable:
    splits = [None] * num_splits
    op_iter = iter(it)
    try:
        while True:
            for n_split in range(num_splits):
                op = next(op_iter)
                if splits[n_split] is None:
                    splits[n_split] = [op]
                else:
                    splits[n_split].append(op)
    except StopIteration:
        pass
    return splits

def schedule_ops(ops_list: List[BaseOperator], num_parallel: int = 1) -> List[BaseOperator]:
    """ Build parallel operator chains. Returns list of last operators in each chain. """

    if not ops_list:
        raise ValueError('Empty list')
    if len(ops_list) == 1:
        return ops_list[0]
    
    splits = split_iterable(ops_list, num_parallel)
    for g in splits: 
        for a, b in pairwise(g):
            a.set_downstream(b)

    return [x[-1] for x in splits]

def is_same_database_uri(uri_1: str, uri_2: str) -> bool:
    """ Compares two database URIs abd returns `True` if they point to the same database.
    `Schema` part of URI (usually encoded as `schema=xxx` query string) is ignored. 
    Note that all URI parts are compared only lexigraphically.
    """
    p1 = urlparse(uri_1)
    p2 = urlparse(uri_2)
    return p1.scheme == p2.scheme and p1.netloc == p2.netloc and p1.username == p2.username and p1.path == p2.path

def adjust_table_name_case(table: BaseTable, db: BaseDatabase) -> BaseTable:

    name = table.name
    schema = table.metadata.schema if table.metadata and table.metadata.schema else None
    
    match db.sql_type:
        case 'postgresql':
            # names must be lower case
            new_name = name.lower() if not name.startswith('"') else name
            new_schema = schema.lower() if schema is not None and not schema.startswith('"') else schema
            return Table(
                new_name,
                conn_id=table.conn_id, 
                metadata=Metadata(new_schema, table.metadata.database if table.metadata else None), 
                columns=table.columns,
                temp=table.temp)

        case 'oracle':
            # names must be upper case
            new_name = name.upper() if not name.startswith('"') else name
            new_schema = schema.upper() if schema is not None and not schema.startswith('"') else schema
            return Table(
                new_name,
                conn_id=table.conn_id, 
                metadata=Metadata(new_schema, table.metadata.database if table.metadata else None), 
                columns=table.columns,
                temp=table.temp)

    return table


# Astro SDK Extras project
# (c) kol, 2023

""" Templated SQL execution """

import logging
from airflow.models.xcom_arg import XComArg
from airflow.models.dag import DagContext
from airflow.utils.task_group import TaskGroup

from astro import sql as aql
from astro.sql.table import Table
from astro.airflow.datasets import kwargs_with_datasets

from typing import Optional, Iterable

from ..utils.utils import schedule_ops
from ..utils.template import get_template

_logger = logging.getLogger('airflow.task')

def run_sql_template(
    template: str,
    conn_id: str,
    input_tables: Iterable[Table] | None = None,
    affected_tables: Iterable[Table] | None = None,
    **kwargs
) -> XComArg:
    """ Runs an SQL script from template file.
    
    SQL script is a single SQL statement or multiple statements 
    to do something good within a single database.

    This function renders given template and executes resulting SQL. Templates are
    text files with optional macros (see `astro_extras.utils.template.get_template_file`).
    Templates are DAG specific.

    Normally, only `INSERT`, `UPDATE` and DDL constructs are used,
    but scope of statements is not limited. However, no data is ever returned to the caller,
    use Astro SDK `run_raw_sql` or `astro_extras.operators.table.load_table` instead if needed.

    Args:
        template: Template name. This should be a base name of `.sql` file
            located at <dags_home>/templates/<dag.dag_id> folder
        conn_id:    Connection to database where to execute templated SQL
        kwargs:     Any extra keyword arguments passed to Astro SDK's `run_raw_sql` function

    Returns:
        Task wrapped in `XComArg`

    Examples:
        This will execute file `./dags/templates/test_dag/test_template.sql` 
        on `target_db` connection:

        >>> with DAG(dag_id='test_dag', ...) as dag:
        >>>     run_sql_template('test_template', 'target_db')
    """

    dag = DagContext.get_current_dag()
    @aql.run_raw_sql(conn_id=conn_id,
                    task_id=f'run-{template}',
                    response_size=0,
                    **kwargs_with_datasets(kwargs=kwargs, 
                               input_datasets=input_tables,
                               output_datasets=affected_tables))
    def _run_sql(template: str):
        sql = get_template(template, '.sql', dag=dag, fail_if_not_found=True)
        _logger.info(f'Executing: {sql}')
        return sql

    return _run_sql(template)

def run_sql_templates(
    templates: Iterable[str],
    conn_id: str,
    group_id: str | None = None,
    num_parallel: int = 1,
    **kwargs,
) -> TaskGroup:
    """ Runs SQL scripts from multiple template files wrapping them up
    in Airflow's `TaskGroup`. See `run_sql_template` for details.

    Args:
        templates: List of template names, which are to be base names of `.sql` files
            located at <dags_home>/templates/<dag.dag_id> folder
        conn_id:    Connection to database where to execute templated SQL
        group_id:   Optional `TaskGroup` id
        num_parallel:   Expected number of parallel task sequences within a group.
            For example, if 4 templates are to be processed, and `num_parallel=2`,
            then 4 resulting tasks will be linked pairwise forming 2 separate chains.
        kwargs:     Any extra keyword arguments passed to Astro SDK's `run_raw_sql` function

    Returns:
        Airflow `TaskGroup`
    """

    with TaskGroup(group_id or 'run-sql', add_suffix_on_collision=True) as tg:
        ops_list = []
        for templ in templates:
            op = run_sql_template(templ, conn_id, **kwargs)
            ops_list.append(op)
        schedule_ops(ops_list, num_parallel)
    return tg



# Astro SDK Extras project
# (c) kol, 2023

""" Timezone conversion routines """

from datetime import datetime, timedelta
from dateutil.tz import tzlocal, tzutc
from dateutil.relativedelta import relativedelta

def localnow() -> datetime:
    """ Return datetime.now() in a local timezone (TZ-aware) """
    return datetime.now(tzlocal())

def datetime_is_tz_aware(dt: datetime) -> bool:
    """ Check datetime is TZ-aware """
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None

def datetime_to_local(dt: datetime) -> datetime:
    """ Convert datetime of any type (TZ-aware or TZ-naive) to local TZ """
    return datetime_to_tz(dt, tzlocal())

def datetime_to_utc(dt: datetime) -> datetime:
    """ Convert datetime of any type (TZ-aware or TZ-naive) to UTC """
    return datetime_to_tz(dt, tzutc())

def datetime_to_tz(dt: datetime, tz=None) -> datetime:
    """ Convert datetime of any type (TZ-aware or TZ-naive) to given timezone """
    if not dt:
        return None
    try:
        return dt.astimezone(tz)
    except (ValueError,TypeError):
        return dt.replace(tzinfo=tz)

def datetime_to_naive(dt: datetime) -> datetime:
    """ Remove TZ component making datetime TZ-naive """
    if not dt:
        return None
    return dt.replace(tzinfo=None)

def days_ago(n_days: int) -> datetime:
    """ Returns TZ-aware date/time `n_days` days in the past from now """
    return localnow() - timedelta(days=n_days)

def months_ago(n_months: int) -> datetime:
    """ Returns TZ-aware date/time `n_months` months in the past from now """
    return localnow() - relativedelta(months=n_months)



# Astro SDK Extras project
# (c) kol, 2023-2024
"""
Setup routine for astro_extras package
"""
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()

setuptools.setup(
    name="astro_extras",
    version="0.1.3",
    author="Kol",
    author_email="skolchin@gmail.com",
    description="Additional Astro SDK operators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/skolchin/astro-extras",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Framework :: Apache Airflow",
        "Framework :: Apache Airflow :: Provider",
    ],
    package_dir={"": "src"},
    packages=['astro_extras', 'astro_extras.hooks', 'astro_extras.operators', 'astro_extras.sensors', 'astro_extras.utils'],
    package_data = {"": ["templates/*.sql"]},
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=requirements,
    setup_requires=['setuptools', 'wheel'],
)
